package hwloc.utils;

import hwloc.Cache;
import hwloc.CacheAggregator;
import hwloc.Core;
import hwloc.Machine;
import hwloc.MachineHierarchicalNode;
import hwloc.PU;
import hwloc.Socket;
import hwloc.UnknownXMLFormatException;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;

/**
 * 2012
 * 
 * @author Emilio Francesquini <francesquini@gmail.com>
 * 
 */
public class DomParser {

	private Machine machine;
	private Document dom;

	public Machine load(InputSource is) {
		parseXmlFile(is);
		parseDocument();
		return machine;
	}

	private void parseXmlFile(InputSource is) {
		// get the factory
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		try {
			DocumentBuilder db = dbf.newDocumentBuilder();
			db.setEntityResolver(new EntityResolver() {
				public InputSource resolveEntity(String publicId,
						String systemId) {
					InputSource ret = null;
					if (systemId.endsWith("hwloc.dtd")) {
						InputStream is = DomParser.class.getClassLoader()
								.getResourceAsStream("hwloc.dtd");
						ret = new InputSource(is);
					}
					return ret;
				}
			});
			dom = db.parse(is);
		} catch (Exception ioe) {
			throw new RuntimeException(ioe);
		}
	}

	private void parseDocument() {
		Element docEle = dom.getDocumentElement();
		NodeList nl = docEle.getChildNodes();
		if (nl != null)
			for (int i = 0; i < nl.getLength(); i++) {				
				Node node = nl.item(i);
				if (node.getNodeType() == Node.ELEMENT_NODE &&  "object".equals(((Element)node).getTagName())) {
					if (machine != null)
						throw new UnknownXMLFormatException();
					machine = getMachineFor((Element)node);
				}
			}
		else
			throw new UnknownXMLFormatException();
	}

	private Machine getMachineFor(Element el) {
		String type = el.getAttribute("type");
		if (!"Machine".equals(type))
			throw new UnknownXMLFormatException("Received type " + type
					+ " when expecting Machine");
		long memory = Long.parseLong(el.getAttribute("local_memory"));
		NodeList nl = el.getChildNodes();
		if (nl != null && nl.getLength() > 0) {
			int length = nl.getLength();
			List<MachineHierarchicalNode> children = new ArrayList<MachineHierarchicalNode>();
			for (int i = 0; i < length; i++) {
				Node child = (Node) nl.item(i);
				if (child.getNodeType() == Node.ELEMENT_NODE) {
					Element childElement = (Element) child;
					if ("object".equals(childElement.getTagName())) {
						type = childElement.getAttribute("type");
						if ("Socket".equals(type)) // TODO colocar numa
							children.add(getSocketFor(childElement));
						else
							throw new UnknownXMLFormatException(
									"Expected Socket but was " + type);
					}
				}
			}
			return new Machine(children.toArray(new Socket[0]), memory);
		}
		throw new UnknownXMLFormatException();
	}

	private Socket getSocketFor(Element el) {
		Socket ret = new Socket();
		int osIndex = getOSIndex(el);
		List<Cache> cacheList = new ArrayList<Cache>();
		NodeList nl = el.getChildNodes();
		int length = nl.getLength();
		for (int i = 0; i < length; i++) {
			Node node = nl.item(i);
			if (node.getNodeType() == Node.ELEMENT_NODE)
				cacheList.add(getCacheFor((Element)node, ret));
		}
		ret.initialize(osIndex, cacheList.toArray(new Cache[0]));
		return ret;
	}

	private int getOSIndex(Element el) {
		return Integer.parseInt(el.getAttribute("os_index"));
	}
	
	private Cache getCacheFor(Element el, CacheAggregator parent) {
		int lineSize = Integer.parseInt(el.getAttribute("cache_linesize"));
		long size = Long.parseLong(el.getAttribute("cache_size"));
		int level = Integer.parseInt(el.getAttribute("depth"));
		Cache ret = new Cache();
		boolean isCache = false, isCore = false;
		//the cache might have another cache or a core beneath it
		NodeList nl = el.getChildNodes();
		int length = nl.getLength();
		List<MachineHierarchicalNode> cacheList = new ArrayList<MachineHierarchicalNode>();
		for (int i = 0; i < length; i++) {
			Node node = nl.item(i);
			if (node.getNodeType() == Node.ELEMENT_NODE) {
				Element childElement = (Element)node;
				String type = childElement.getAttribute("type");
				if ("Cache".equals(type)) {
					cacheList.add(getCacheFor(childElement, ret));
					isCache = true;
				}
				else if ("Core".equals(type)) {
					cacheList.add(getCoreFor(childElement, ret));
					isCore = true;
				}
				else
					throw new UnknownXMLFormatException("Expected Cache or Core, got: " + type);
			}
		}
		if (! (isCache ^ isCore))
			throw new UnknownXMLFormatException("Expected Cache or Core");
		Core[] cores = isCore ? cacheList.toArray(new Core[0]) : null;
		Cache[] caches = isCache ? cacheList.toArray(new Cache[0]) : null;
		ret.initialize(parent, size, lineSize, level, cores, caches);
		return ret;
	}

	private MachineHierarchicalNode getCoreFor(Element el, MachineHierarchicalNode parent) {
		Core ret = new Core();
		int osIndex = getOSIndex(el);
		List<PU> pus = new ArrayList<PU>();
		NodeList nl = el.getChildNodes();
		int length = nl.getLength();
		for (int i = 0; i < length; i++) {
			Node node = nl.item(i);
			if (node.getNodeType() == Node.ELEMENT_NODE) {
				Element childElement = (Element)node;
				String type = childElement.getAttribute("type");
				if ("PU".equals(type))
					pus.add(getPUFor(childElement, ret));
				else
					throw new UnknownXMLFormatException("Expected Cache or Core, got: " + type);
			}
		}		
		ret.initialize(osIndex, parent, pus.toArray(new PU[0]));
		return ret;
	}

	
	private PU getPUFor(Element el, Core parent) {
		return new PU(getOSIndex(el), parent);
	}

}
